/* Virtual Reality: MFunc.cpp
   =============================================================================

   This file contains the functions for the MFunc type. The MFunc type is the
   internal representation of a function in a N64 program. Such a function has
   a name, offset and a length. For recompiling purposes it also has a checksum
   and the address where the function is in memory.

   12-12-2000 First release
*/

#include <stdio.h>

#include "MFunc.h"
#include "MFuncGetName.h"

#define Max(x, y) (x > y) ? (x) : (y)

uWord MFunc_GetLength(MFunc Func)
{
	MOffset MaxPc     = Func.Offset;
	MOffset CurrentPc = Func.Offset - 4;

	do
	{
		CurrentPc += 4;

		MInstr Instr = MInstr_Build(CurrentPc, ((uWord *)Func.Source)[(CurrentPc - Func.Offset) / 4]);

		if (MInstr_IsBranch(Instr)) 
		{
			MaxPc = Max(MaxPc, MInstr_BranchPc(Instr));
			MaxPc = Max(MaxPc, CurrentPc + 4);
		}
		else if (MInstr_IsLikelyBranch(Instr))
		{
			MaxPc = Max(MaxPc, MInstr_BranchPc(Instr));
			MaxPc = Max(MaxPc, CurrentPc + 8);
		}
		else if (MInstr_IsJump(Instr))
		{
			MaxPc = Max(MaxPc, CurrentPc + 4);
			CurrentPc += 4;
		}
		else if (!MInstr_IsBreak(Instr))
			MaxPc = Max(MaxPc, CurrentPc + 4);
	} while (CurrentPc < MaxPc);

	return (MaxPc - Func.Offset + 4) / 4;
}

uWord MFunc_GetChecksum(MFunc Func)
{
	uWord Checksum = 0;
	uWord *InstrAddress = (uWord *)Func.Source;

	for (int InstrNr = 0; InstrNr < Func.Length; InstrNr++)
		Checksum += InstrAddress[InstrNr]; 

	return Checksum;
}

MList MFunc_GetInstrs(MFunc Func)
{
	uWord *InstrAddress = (uWord *)Func.Source;
	MList List = MList_Empty();

	for (int InstrNr = 0; InstrNr < Func.Length; InstrNr++)
	{
		MInstr Instr = MInstr_Build(Func.Offset + 4 * InstrNr, InstrAddress[InstrNr]);
		List = MList_Add(List, Instr);
	}

	return List;
}

MFunc MFunc_GetFunction(MOffset Offset, char *Source)
{
	MFunc Func;

	Func.Source   = Source;
	Func.Offset   = Offset;
	Func.Length   = MFunc_GetLength(Func);
	Func.Checksum = MFunc_GetChecksum(Func);
	Func.Instrs   = MFunc_GetInstrs(Func);
	Func.Name     = MFunc_GetName(Func);

	return Func;
}

void MFunc_Print(MFunc Func)
{
	printf("============================================================================================\n");
	printf("MipsFunction: %s, starting at offset: 0x%08x, length: %d instrs, checksum: %08x\n", Func.Name, Func.Offset, Func.Length, Func.Checksum);
	printf("============================================================================================\n\n");
	MList_Print(Func.Instrs);
}
