/* Virtual Reality: MInstr.cpp
   =============================================================================

   This file contains the functions for the MInstr type. The MInstr type is the
   internal representation of a Mips assembly instruction.

   12-12-2000 First release
*/

#include "MInstr.h"

MInstr MInstr_Build(MOffset Offset, uWord Instr)
{
	MInstr NewInstr;

	NewInstr.Offset = Offset;
	NewInstr.Instr  = Instr;

	return NewInstr;
}

uByte MInstr_Opcode(MInstr Instr)
{
	return ((uByte)((Instr.Instr >> 26) & 0x002f));
}

uByte MInstr_Sa(MInstr Instr)
{
	return ((uByte)((Instr.Instr >>  6) & 0x001f));
}

MReg MInstr_Rs(MInstr Instr)
{
	return ((uByte)((Instr.Instr >> 21) & 0x001f));
}

MReg MInstr_Rt(MInstr Instr)
{
	return ((uByte)((Instr.Instr >> 16) & 0x001f));
}

MReg MInstr_Rd(MInstr Instr)
{
	return ((uByte)((Instr.Instr >> 11) & 0x001f));
}

MReg MInstr_Fs(MInstr Instr)
{
	return ((uByte)((Instr.Instr >> 11) & 0x001f));
}

MReg MInstr_Ft(MInstr Instr)
{
	return ((uByte)((Instr.Instr >> 16) & 0x001f));
}

MReg MInstr_Fd(MInstr Instr)
{
	return ((uByte)((Instr.Instr >> 6) & 0x001f));
}

uByte MInstr_Function(MInstr Instr)
{
	return ((uByte)((Instr.Instr) & 0x003f));
}

uByte MInstr_Fmt(MInstr Instr)
{
	return ((uByte)((Instr.Instr >> 21) & 0x001f));
}

uByte MInstr_Nd(MInstr Instr)
{
	return ((uByte)((Instr.Instr >> 17) & 0x0001));
}

uHword MInstr_Immediate(MInstr Instr)
{
	return ((uHword)((Instr.Instr) & 0xffff));
}

MOffset MInstr_BranchPc(MInstr Instr)
{
	return ((MOffset)(Instr.Offset + 4 + ((sHword)Instr.Instr << 2)));
}

MOffset MInstr_JumpPc(MInstr Instr)
{
	return ((MOffset)((Instr.Offset & 0xf0000000) | ((Instr.Instr & 0x03ffffff) << 2)));
}

bool MInstr_IsBranch(MInstr Instr)
{
	switch (MInstr_Opcode(Instr))
	{
		case OP_BEQ  :
		case OP_BNE  :
		case OP_BLEZ :
		case OP_BGTZ :
			return true;
		case OP_REGIMM :
			switch (MInstr_Rt(Instr))
			{
				case RG_BLTZ   :
				case RG_BGEZ   :
				case RG_BLTZAL : 
				case RG_BGEZAL : 
					return true;
				default :
					return false;
			}
		case OP_COP1 :
			return (MInstr_Fmt(Instr) == COP1_BC1 && !MInstr_Nd(Instr));
		default :
			return false;
	}
}

bool MInstr_IsLikelyBranch(MInstr Instr)
{
	switch (MInstr_Opcode(Instr))
	{
		case OP_BEQL  :
		case OP_BNEL  :
		case OP_BLEZL :
		case OP_BGTZL :
			return true;
		case OP_REGIMM :
			switch (MInstr_Rt(Instr))
			{
				case RG_BLTZL   :
				case RG_BGEZL   :
				case RG_BLTZALL : 
				case RG_BGEZALL : 
					return true;
				default :
					return false;
			}
		case OP_COP1 :
			return (MInstr_Fmt(Instr) == COP1_BC1 && MInstr_Nd(Instr));
		default :
			return false;
	}
}

bool MInstr_IsCall(MInstr Instr)
{
	switch (MInstr_Opcode(Instr))
	{
		case OP_JAL : 
			return true;
		case OP_SPECIAL :
			switch (MInstr_Function(Instr))
			{
				case SP_JALR :
					return true;
				default :
					return false;
			}
		default :
			return false;
	}
}

bool MInstr_IsBreak(MInstr Instr)
{
	return (MInstr_Opcode(Instr) == OP_SPECIAL && MInstr_Function(Instr) == SP_BREAK);
}

bool MInstr_IsJump(MInstr Instr)
{
	switch (MInstr_Opcode(Instr))
	{
		case OP_J : 
			return true;
		case OP_SPECIAL :
			switch (MInstr_Function(Instr))
			{
				case SP_JR :
					return true;
				default :
					return false;
			}
		default :
			return false;
	}
}
