#ifndef H_MInstr
#define H_MInstr

#include "MTypes.h"

typedef uByte MReg;    
typedef uWord MOffset;

typedef struct
{
	MOffset Offset;
	uWord   Instr;
} MInstr;              


MInstr MInstr_Build          (MOffset, uWord); 

uByte  MInstr_Opcode         (MInstr); 
uByte  MInstr_Sa             (MInstr); 
uByte  MInstr_Rs             (MInstr);
uByte  MInstr_Rt             (MInstr);
uByte  MInstr_Rd             (MInstr);
uByte  MInstr_Fs             (MInstr);
uByte  MInstr_Ft             (MInstr);
uByte  MInstr_Fd             (MInstr);
uByte  MInstr_Function       (MInstr);
uByte  MInstr_Fmt            (MInstr);
uByte  MInstr_Nd             (MInstr);
uHword MInstr_Immediate      (MInstr);

uWord  MInstr_BranchPc       (MInstr);
uWord  MInstr_JumpPc         (MInstr);

bool   MInstr_IsBranch       (MInstr);
bool   MInstr_IsLikelyBranch (MInstr);
bool   MInstr_IsCall         (MInstr);
bool   MInstr_IsBreak        (MInstr);
bool   MInstr_IsJump         (MInstr);

#define OP_SPECIAL  0x00
#define OP_REGIMM   0x01
#define OP_J        0x02
#define OP_JAL      0x03
#define OP_BEQ      0x04
#define OP_BNE      0x05
#define OP_BLEZ     0x06
#define OP_BGTZ     0x07
#define OP_ADDI     0x08
#define OP_ADDIU    0x09
#define OP_SLTI     0x0A
#define OP_SLTIU    0x0B
#define OP_ANDI     0x0C
#define OP_ORI      0x0D
#define OP_XORI     0x0E
#define OP_LUI      0x0F
#define OP_COP0     0x10
#define OP_COP1     0x11
#define OP_BEQL     0x14
#define OP_BNEL     0x15
#define OP_BLEZL    0x16
#define OP_BGTZL    0x17
#define OP_DADDI    0x18
#define OP_DADDIU   0x19
#define OP_LDL      0x1A
#define OP_LDR      0x1B
#define OP_LB       0x20
#define OP_LH       0x21
#define OP_LWL      0x22
#define OP_LW       0x23
#define OP_LBU      0x24
#define OP_LHU      0x25
#define OP_LWR      0x26
#define OP_LWU      0x27
#define OP_SB       0x28
#define OP_SH       0x29
#define OP_SWL      0x2A
#define OP_SW       0x2B
#define OP_SDL      0x2C
#define OP_SDR      0x2D
#define OP_SWR      0x2E
#define OP_CACHE    0x2F
#define OP_LL       0x30
#define OP_LWC1     0x31
#define OP_LLD      0x34
#define OP_LDC1     0x35  
#define OP_LDC2     0x36
#define OP_LD       0x37
#define OP_SC       0x38
#define OP_SWC1     0x39
#define OP_SCD      0x3C
#define OP_SDC1     0x3D
#define OP_SDC2     0x3E
#define OP_SD       0x3F

#define SP_SLL      0x00 
#define SP_SRL      0x02
#define SP_SRA      0x03
#define SP_SLLV     0x04
#define SP_SRLV     0x06
#define SP_SRAV     0x07
#define SP_JR       0x08
#define SP_JALR     0x09
#define SP_SYSCALL  0x0C
#define SP_BREAK    0x0D
#define SP_SYNC     0x0F
#define SP_MFHI     0x10
#define SP_MTHI     0x11
#define SP_MFLO     0x12
#define SP_MTLO     0x13
#define SP_DSLLV    0x14
#define SP_DSRLV    0x16
#define SP_DSRAV    0x17
#define SP_MULT     0x18
#define SP_MULTU    0x19
#define SP_DIV      0x1A
#define SP_DIVU     0x1B
#define SP_DMULT    0x1C
#define SP_DMULTU   0x1D
#define SP_DDIV     0x1E
#define SP_DDIVU    0x1F
#define SP_ADD      0x20
#define SP_ADDU     0x21
#define SP_SUB      0x22
#define SP_SUBU     0x23
#define SP_AND      0x24
#define SP_OR       0x25
#define SP_XOR      0x26
#define SP_NOR      0x27
#define SP_SLT      0x2A
#define SP_SLTU     0x2B
#define SP_DADD     0x2C
#define SP_DADDU    0x2D
#define SP_DSUB     0x2E
#define SP_DSUBU    0x2F
#define SP_TGE      0x30
#define SP_TGEU     0x31
#define SP_TLT      0x32
#define SP_TLTU     0x33
#define SP_TEQ      0x34
#define SP_TNE      0x36
#define SP_DSLL     0x38
#define SP_DSRL     0x3A
#define SP_DSRA     0x3B
#define SP_DSLL32   0x3C
#define SP_DSRL32   0x3E
#define SP_DSRA32   0x3F

#define RG_BLTZ     0x00
#define RG_BGEZ     0x01
#define RG_BLTZL    0x02
#define RG_BGEZL    0x03
#define RG_TGEI     0x08
#define RG_TGEIU    0x09
#define RG_TLTI     0x0A
#define RG_TLTIU    0x0B
#define RG_TEQI     0x0C
#define RG_TNEI     0x0E
#define RG_BLTZAL   0x10
#define RG_BGEZAL   0x11
#define RG_BLTZALL  0x12
#define RG_BGEZALL  0x13

#define COP0_MFC0   0x00
#define COP0_MTC0   0x04
#define COP0_TLB    0x10

#define TLB_TLBR    0x01
#define TLB_TLBWI   0x02
#define TLB_TLBWR   0x06
#define TLB_TLBP    0x08
#define TLB_ERET    0x18

#define COP1_MFC1   0x00
#define COP1_DMFC1  0x01
#define COP1_CFC1   0x02
#define COP1_MTC1   0x04
#define COP1_DMTC1  0x05
#define COP1_CTC1   0x06
#define COP1_BC1    0x08
#define COP1_S      0x10
#define COP1_D      0x11
#define COP1_W      0x14
#define COP1_L      0x15

#define BC1_BC1F    0x00
#define BC1_BC1T    0x01
#define BC1_BC1FL   0x02
#define BC1_BC1TL   0x03

#define FP_ADD       0x00
#define FP_SUB       0x01
#define FP_MUL       0x02
#define FP_DIV       0x03
#define FP_SQRT      0x04
#define FP_ABS       0x05
#define FP_MOV       0x06
#define FP_NEG       0x07
#define FP_ROUNDL    0x08
#define FP_TRUNCL    0x09
#define FP_CEILL     0x0A
#define FP_FLOORL    0x0B
#define FP_ROUNDW    0x0C
#define FP_TRUNCW    0x0D
#define FP_CEILW     0x0E
#define FP_FLOORW    0x0F
#define FP_CVTS      0x20
#define FP_CVTD      0x21
#define FP_CVTW      0x24
#define FP_CVTL      0x25
#define FP_CF        0x30
#define FP_CUN       0x31
#define FP_CEQ       0x32
#define FP_CUEQ      0x33
#define FP_COLT      0x34
#define FP_CULT      0x35
#define FP_COLE      0x36
#define FP_CULE      0x37
#define FP_CSF       0x38
#define FP_CNGLE     0x39
#define FP_CSEQ      0x3A
#define FP_CNGL      0x3B
#define FP_CLT       0x3C
#define FP_CNGE      0x3D
#define FP_CLE       0x3E
#define FP_CNGT      0x3F

#endif
