/* Virtual Reality: MInstrToString.cpp
   =============================================================================

   This file contains the functions to convert a MInstr into a readable String.

   12-12-2000 First release
*/

#include <stdio.h>

#include "MInstrToString.h"

#define Rs         MInstr_CpuRegToString(MInstr_Rs(Instr))
#define Rt         MInstr_CpuRegToString(MInstr_Rt(Instr))
#define Rd         MInstr_CpuRegToString(MInstr_Rd(Instr))
#define Fs         MInstr_FpuRegToString(MInstr_Fs(Instr))
#define Ft         MInstr_FpuRegToString(MInstr_Ft(Instr))
#define Fd         MInstr_FpuRegToString(MInstr_Fd(Instr))
#define Cop0Rd     MInstr_Cop0RegToString(MInstr_Rd(Instr))
#define Immediate  MInstr_Immediate(Instr)
#define Sa         MInstr_Sa(Instr)
#define JumpPc     MInstr_JumpPc(Instr)
#define BranchPc   MInstr_BranchPc(Instr)
#define Offset     Instr.Offset

char *MInstr_CpuRegToString(uByte RegNr)
{
	switch(RegNr)
	{
		case 0x00 : return "r0";
		case 0x01 : return "at";
		case 0x02 : return "v0";
		case 0x03 : return "v1";
		case 0x04 : return "a0";
		case 0x05 : return "a1";
		case 0x06 : return "a2";
		case 0x07 : return "a3";
		case 0x08 : return "t0";
		case 0x09 : return "t1";
		case 0x0a : return "t2";
		case 0x0b : return "t3";
		case 0x0c : return "t4";
		case 0x0d : return "t5";
		case 0x0e : return "t6";
		case 0x0f : return "t7";
		case 0x10 : return "s0";
		case 0x11 : return "s1";
		case 0x12 : return "s2";
		case 0x13 : return "s3";
		case 0x14 : return "s4";
		case 0x15 : return "s5";
		case 0x16 : return "s6";
		case 0x17 : return "s7";
		case 0x18 : return "t8";
		case 0x19 : return "t9";
		case 0x1a : return "k0";
		case 0x1b : return "k1";
		case 0x1c : return "gp";
		case 0x1d : return "sp";
		case 0x1e : return "s8";
		case 0x1f : return "ra";
		default   : return "xx";
	}
}

char *MInstr_Cop0RegToString(uByte RegNr)
{
	switch(RegNr)
	{
		case 0x00 : return "index";
		case 0x01 : return "random";
		case 0x02 : return "entrylo0";
		case 0x03 : return "entrylo1";
		case 0x04 : return "context";
		case 0x05 : return "pagemask";
		case 0x06 : return "wird";
		case 0x07 : return "reserved";
		case 0x08 : return "badvaddr";
		case 0x09 : return "count";
		case 0x0a : return "entryhi";
		case 0x0b : return "compare";
		case 0x0c : return "status";
		case 0x0d : return "cause";
		case 0x0e : return "epc";
		case 0x0f : return "previd";
		case 0x10 : return "config";
		case 0x11 : return "lladdr";
		case 0x12 : return "watchlo";
		case 0x13 : return "watchhi";
		case 0x14 : return "xcontext";
		case 0x15 : return "reserved";
		case 0x16 : return "reserved";
		case 0x17 : return "reserved";
		case 0x18 : return "reserved";
		case 0x19 : return "reserved";
		case 0x1a : return "perr";
		case 0x1b : return "cacheerr";
		case 0x1c : return "taglo";
		case 0x1d : return "taghi";
		case 0x1e : return "errorepc";
		case 0x1f : return "reserved";
		default   : return "xxxxxxxx";
	}
}

char *MInstr_FpuRegToString(uByte RegNr)
{
	switch(RegNr)
	{
		case 0x00 : return "f0";
		case 0x01 : return "f1";
		case 0x02 : return "f2";
		case 0x03 : return "f3";
		case 0x04 : return "f4";
		case 0x05 : return "f5";
		case 0x06 : return "f6";
		case 0x07 : return "f7";
		case 0x08 : return "f8";
		case 0x09 : return "f9";
		case 0x0a : return "f10";
		case 0x0b : return "f11";
		case 0x0c : return "f12";
		case 0x0d : return "f13";
		case 0x0e : return "f14";
		case 0x0f : return "f15";
		case 0x10 : return "f16";
		case 0x11 : return "f17";
		case 0x12 : return "f18";
		case 0x13 : return "f19";
		case 0x14 : return "f20";
		case 0x15 : return "f21";
		case 0x16 : return "f22";
		case 0x17 : return "f23";
		case 0x18 : return "f24";
		case 0x19 : return "f25";
		case 0x1a : return "f26";
		case 0x1b : return "f27";
		case 0x1c : return "f28";
		case 0x1d : return "f29";
		case 0x1e : return "f30";
		case 0x1f : return "f31";
		default   : return "xx";
	}
}

char *MInstr_ToString_Special(MInstr Instr)
{
	static char String[50];

	switch (MInstr_Function(Instr))
	{
		case SP_SLL     : 
			if (Instr.Instr)
				sprintf(String, "%08x sll      %s = %s << %02x", Offset, Rd, Rt, Sa);
			else
				sprintf(String, "%08x nop", Offset);
			break;
		case SP_SRL     : sprintf(String, "%08x srl      %s = %s >> %02x", Offset, Rd, Rt, Sa); break; 
		case SP_SRA     : sprintf(String, "%08x sra      %s = %s >> %02x", Offset, Rd, Rt, Sa); break; 
		case SP_SLLV    : sprintf(String, "%08x sllv     %s = %s << %s", Offset, Rd, Rt, Rs); break; 
		case SP_SRLV    : sprintf(String, "%08x srlv     %s = %s >> %s", Offset, Rd, Rt, Rs); break; 
		case SP_SRAV    : sprintf(String, "%08x srav     %s = %s >> %s", Offset, Rd, Rt, Rs); break; 
		case SP_JR      : sprintf(String, "%08x jr       ->%s", Offset, Rs); break;
		case SP_JALR    : sprintf(String, "%08x jalr     (%s = %08x) ->%s", Offset, Rd, Offset + 8, Rs); break;
		case SP_SYSCALL : sprintf(String, "%08x syscall", Offset); break;
		case SP_BREAK   : sprintf(String, "%08x break", Offset); break;
		case SP_SYNC    : sprintf(String, "%08x sync", Offset); break;
		case SP_MFHI    : sprintf(String, "%08x mfhi     %s = hi", Offset, Rd); break;
		case SP_MTHI    : sprintf(String, "%08x mthi     hi = %s", Offset, Rs); break;
		case SP_MFLO    : sprintf(String, "%08x mflo     %s = lo", Offset, Rd); break;
		case SP_MTLO    : sprintf(String, "%08x mtlo     lo = %s", Offset, Rs); break;
		case SP_DSLLV   : sprintf(String, "%08x dsllv    %s = %s << %s", Offset, Rd, Rt, Rs); break; 
		case SP_DSRLV   : sprintf(String, "%08x dsrlv    %s = %s >> %s", Offset, Rd, Rt, Rs); break; 
		case SP_DSRAV   : sprintf(String, "%08x dsrav    %s = %s >> %s", Offset, Rd, Rt, Rs); break; 
		case SP_MULT    : sprintf(String, "%08x mult     (lo, hi) = %s * %s", Offset, Rs, Rt); break;
		case SP_MULTU   : sprintf(String, "%08x multu    (lo, hi) = %s * %s", Offset, Rs, Rt); break;
		case SP_DIV     : sprintf(String, "%08x div      (lo, hi) = %s / %s", Offset, Rs, Rt); break;
		case SP_DIVU    : sprintf(String, "%08x divu     (lo, hi) = %s / %s", Offset, Rs, Rt); break;
		case SP_DMULT   : sprintf(String, "%08x dmult    (lo, hi) = %s * %s", Offset, Rs, Rt); break;
		case SP_DMULTU  : sprintf(String, "%08x dmultu   (lo, hi) = %s * %s", Offset, Rs, Rt); break;
		case SP_DDIV    : sprintf(String, "%08x ddiv     (lo, hi) = %s / %s", Offset, Rs, Rt); break;
		case SP_DDIVU   : sprintf(String, "%08x ddivu    (lo, hi) = %s / %s", Offset, Rs, Rt); break;
		case SP_ADD     : sprintf(String, "%08x add      %s = %s + %s", Offset, Rd, Rs, Rt); break;
		case SP_ADDU    : sprintf(String, "%08x addu     %s = %s + %s", Offset, Rd, Rs, Rt); break;
		case SP_SUB     : sprintf(String, "%08x sub      %s = %s - %s", Offset, Rd, Rs, Rt); break;
		case SP_SUBU    : sprintf(String, "%08x subu     %s = %s - %s", Offset, Rd, Rs, Rt); break;
		case SP_AND     : sprintf(String, "%08x and      %s = %s & %s", Offset, Rd, Rs, Rt); break;
		case SP_OR      : sprintf(String, "%08x or       %s = %s | %s", Offset, Rd, Rs, Rt); break;
		case SP_XOR     : sprintf(String, "%08x xor      %s = %s ^ %s", Offset, Rd, Rs, Rt); break;
		case SP_NOR     : sprintf(String, "%08x nor      %s = ~(%s | %s)", Offset, Rd, Rs, Rt); break;
		case SP_SLT     : sprintf(String, "%08x slt      %s = (%s < %s)", Offset, Rd, Rs, Rt); break;
		case SP_SLTU    : sprintf(String, "%08x sltu     %s = (%s < %s)", Offset, Rd, Rs, Rt); break;
		case SP_DADD    : sprintf(String, "%08x dadd     %s = %s + %s", Offset, Rd, Rs, Rt); break;
		case SP_DADDU   : sprintf(String, "%08x daddu    %s = %s + %s", Offset, Rd, Rs, Rt); break;
		case SP_DSUB    : sprintf(String, "%08x dsub     %s = %s - %s", Offset, Rd, Rs, Rt); break;
		case SP_DSUBU   : sprintf(String, "%08x dsubu    %s = %s - %s", Offset, Rd, Rs, Rt); break;
		case SP_TGE     : sprintf(String, "%08x tge      %s >= %s", Offset, Rs, Rt); break;
		case SP_TGEU    : sprintf(String, "%08x tgeu     %s >= %s", Offset, Rs, Rt); break;
		case SP_TLT     : sprintf(String, "%08x tlt      %s < %s", Offset, Rs, Rt); break;
		case SP_TLTU    : sprintf(String, "%08x tltu     %s < %s", Offset, Rs, Rt); break;
		case SP_TEQ     : sprintf(String, "%08x teq      %s == %s", Offset, Rs, Rt); break;
		case SP_TNE     : sprintf(String, "%08x tne      %s != %s", Offset, Rs, Rt); break;
		case SP_DSLL    : sprintf(String, "%08x dsll     %s = %s << %02x", Offset, Rd, Rt, Sa); break;
		case SP_DSRL    : sprintf(String, "%08x dsrl     %s = %s >> %02x", Offset, Rd, Rt, Sa); break;
		case SP_DSRA    : sprintf(String, "%08x dsra     %s = %s >> %02x", Offset, Rd, Rt, Sa); break;
		case SP_DSLL32  : sprintf(String, "%08x dsll32   %s = %s << (32 + %02x)", Offset, Rd, Rt, Sa); break;
		case SP_DSRL32  : sprintf(String, "%08x dsrl32   %s = %s >> (32 + %02x)", Offset, Rd, Rt, Sa); break;
		case SP_DSRA32  : sprintf(String, "%08x dsra32   %s = %s >> (32 + %02x)", Offset, Rd, Rt, Sa); break;
		default         : sprintf(String, "%08x Illegal Instruction", Offset); break;
	}

	return String;
}

char *MInstr_ToString_Regimm(MInstr Instr)
{
	static char String[50];

	switch (MInstr_Rt(Instr))
	{
		case RG_BLTZ    : sprintf(String, "%08x bltz     (%s < 0) ->%08x", Offset, Rs, BranchPc); break;
		case RG_BGEZ    : sprintf(String, "%08x bgez     (%s >= 0) ->%08x", Offset, Rs, BranchPc); break;
		case RG_BLTZL   : sprintf(String, "%08x bltzl    (%s < 0) ->%08x", Offset, Rs, BranchPc); break;
		case RG_BGEZL   : sprintf(String, "%08x bgezl    (%s >= 0) ->%08x", Offset, Rs, BranchPc); break;
		case RG_TGEI    : sprintf(String, "%08x tgei     %s >= %04x", Offset, Rs, Immediate); break;
		case RG_TGEIU   : sprintf(String, "%08x tgeiu    %s >= %04x", Offset, Rs, Immediate); break;
		case RG_TLTI    : sprintf(String, "%08x tlti     %s < %04x", Offset, Rs, Immediate); break;
		case RG_TLTIU   : sprintf(String, "%08x tltiu    %s < %04x", Offset, Rs, Immediate); break;
		case RG_TEQI    : sprintf(String, "%08x teqi     %s == %04x", Offset, Rs, Immediate); break;
		case RG_TNEI    : sprintf(String, "%08x tnei     %s != %04x", Offset, Rs, Immediate); break;
		case RG_BLTZAL  : sprintf(String, "%08x bltzal   (%s < 0) ->%08x", Offset, Rs, BranchPc); break;
		case RG_BGEZAL  : sprintf(String, "%08x bgezal   (%s < 0) ->%08x", Offset, Rs, BranchPc); break;
		case RG_BLTZALL : sprintf(String, "%08x bltzall  (%s < 0) ->%08x", Offset, Rs, BranchPc); break;
		case RG_BGEZALL : sprintf(String, "%08x bgezall  (%s < 0) ->%08x", Offset, Rs, BranchPc); break;
		default         : sprintf(String, "%08x Illegal Instruction", Offset); break;
	}

	return String;
}

char *MInstr_ToString_Tlb(MInstr Instr)
{
	static char String[50];

	switch (MInstr_Function(Instr))
	{
		case TLB_TLBR   : sprintf(String, "%08x tlbr", Offset); break;
		case TLB_TLBWI  : sprintf(String, "%08x tlbwi", Offset); break;
		case TLB_TLBWR  : sprintf(String, "%08x tlbwr", Offset); break;
		case TLB_TLBP   : sprintf(String, "%08x tlbp", Offset); break;
		case TLB_ERET   : sprintf(String, "%08x eret", Offset); break;
		default         : sprintf(String, "%08x Illegal Instruction", Offset); break;
	}

	return String;
}

char *MInstr_ToString_Cop0(MInstr Instr)
{
	static char String[50];

	switch (MInstr_Fmt(Instr))
	{
		case COP0_MFC0  : sprintf(String, "%08x mfc0     %s = %s", Offset, Rt, Cop0Rd); break;
		case COP0_MTC0  : sprintf(String, "%08x mtc0     %s = %s", Offset, Cop0Rd, Rt); break;
		case COP0_TLB   : return MInstr_ToString_Tlb(Instr);
		default         : sprintf(String, "%08x Illegal Instruction", Offset); break;
	}

	return String;
}

char *MInstr_ToString_Bc1(MInstr Instr)
{
	static char String[50];

	switch (MInstr_Rt(Instr))
	{
		case BC1_BC1F   : sprintf(String, "%08x bc1f     c ->%08x", Offset, BranchPc); break;
		case BC1_BC1T   : sprintf(String, "%08x bc1t     c ->%08x", Offset, BranchPc); break;
		case BC1_BC1FL  : sprintf(String, "%08x bc1fl    c ->%08x", Offset, BranchPc); break;
		case BC1_BC1TL  : sprintf(String, "%08x bc1tl    c ->%08x", Offset, BranchPc); break;
		default         : sprintf(String, "%08x Illegal Instruction", Offset); break;
	}

	return String;
}

char *MInstr_ToString_S(MInstr Instr)
{
	static char String[50];

	switch (MInstr_Function(Instr))
	{	
		case FP_ADD      : sprintf(String, "%08x s.add    %s = %s + %s", Offset, Fd, Fs, Ft); break;
		case FP_SUB      : sprintf(String, "%08x s.sub    %s = %s - %s", Offset, Fd, Fs, Ft); break;
		case FP_MUL      : sprintf(String, "%08x s.mul    %s = %s * %s", Offset, Fd, Fs, Ft); break;
		case FP_DIV      : sprintf(String, "%08x s.div    %s = %s / %s", Offset, Fd, Fs, Ft); break;
		case FP_SQRT     : sprintf(String, "%08x s.sqrt   %s = sqrt %s", Offset, Fd, Fs); break;
		case FP_ABS      : sprintf(String, "%08x s.abs    %s = abs %s", Offset, Fd, Fs); break;
		case FP_MOV      : sprintf(String, "%08x s.mov    %s = %s", Offset, Fd, Fs); break;
		case FP_NEG      : sprintf(String, "%08x s.neg    %s = -%s", Offset, Fd, Fs); break;
		case FP_ROUNDL   : sprintf(String, "%08x s.roundl %s = %s", Offset, Fd, Fs); break;
		case FP_TRUNCL   : sprintf(String, "%08x s.truncl %s = %s", Offset, Fd, Fs); break;
		case FP_CEILL    : sprintf(String, "%08x s.ceill  %s = %s", Offset, Fd, Fs); break;
		case FP_FLOORL   : sprintf(String, "%08x s.floorl %s = %s", Offset, Fd, Fs); break;
		case FP_ROUNDW   : sprintf(String, "%08x s.roundw %s = %s", Offset, Fd, Fs); break;
		case FP_TRUNCW   : sprintf(String, "%08x s.truncw %s = %s", Offset, Fd, Fs); break;
		case FP_CEILW    : sprintf(String, "%08x s.ceilw  %s = %s", Offset, Fd, Fs); break;
		case FP_FLOORW   : sprintf(String, "%08x s.floorw %s = %s", Offset, Fd, Fs); break;
		case FP_CVTD     : sprintf(String, "%08x s.cvtd   %s = %s", Offset, Fd, Fs); break;
		case FP_CVTW     : sprintf(String, "%08x s.cvtw   %s = %s", Offset, Fd, Fs); break;
		case FP_CVTL     : sprintf(String, "%08x s.cvtl   %s = %s", Offset, Fd, Fs); break;
		case FP_CF       : sprintf(String, "%08x s.cf     c = %s f %s", Offset, Fs, Ft); break;
		case FP_CUN      : sprintf(String, "%08x s.cun    c = %s un %s", Offset, Fs, Ft); break;
		case FP_CEQ      : sprintf(String, "%08x s.ceq    c = %s eq %s", Offset, Fs, Ft); break;
		case FP_CUEQ     : sprintf(String, "%08x s.cueq   c = %s ueq %s", Offset, Fs, Ft); break;
		case FP_COLT     : sprintf(String, "%08x s.cotl   c = %s olt %s", Offset, Fs, Ft); break;
		case FP_CULT     : sprintf(String, "%08x s.cult   c = %s ult %s", Offset, Fs, Ft); break;
		case FP_COLE     : sprintf(String, "%08x s.cole   c = %s ole %s", Offset, Fs, Ft); break;
		case FP_CULE     : sprintf(String, "%08x s.cule   c = %s ule %s", Offset, Fs, Ft); break;
		case FP_CSF      : sprintf(String, "%08x s.csf    c = %s sf %s", Offset, Fs, Ft); break;
		case FP_CNGLE    : sprintf(String, "%08x s.cngle  c = %s ngle %s", Offset, Fs, Ft); break;
		case FP_CSEQ     : sprintf(String, "%08x s.cseq   c = %s seq %s", Offset, Fs, Ft); break;
		case FP_CNGL     : sprintf(String, "%08x s.cngl   c = %s ngl %s", Offset, Fs, Ft); break;
		case FP_CLT      : sprintf(String, "%08x s.clt    c = %s lt %s", Offset, Fs, Ft); break;
		case FP_CNGE     : sprintf(String, "%08x s.cnge   c = %s nge %s", Offset, Fs, Ft); break;
		case FP_CLE      : sprintf(String, "%08x s.cf     c = %s le %s", Offset, Fs, Ft); break;
		case FP_CNGT     : sprintf(String, "%08x s.cf     c = %s ngt %s", Offset, Fs, Ft); break;
		default          : sprintf(String, "%08x Illegal Instruction", Offset); break;
	}

	return String;
}

char *MInstr_ToString_D(MInstr Instr)
{
	static char String[50];

	switch (MInstr_Function(Instr))
	{	
		case FP_ADD      : sprintf(String, "%08x d.add    %s = %s + %s", Offset, Fd, Fs, Ft); break;
		case FP_SUB      : sprintf(String, "%08x d.sub    %s = %s - %s", Offset, Fd, Fs, Ft); break;
		case FP_MUL      : sprintf(String, "%08x d.mul    %s = %s * %s", Offset, Fd, Fs, Ft); break;
		case FP_DIV      : sprintf(String, "%08x d.div    %s = %s / %s", Offset, Fd, Fs, Ft); break;
		case FP_SQRT     : sprintf(String, "%08x d.sqrt   %s = sqrt %s", Offset, Fd, Fs); break;
		case FP_ABS      : sprintf(String, "%08x d.abs    %s = abs %s", Offset, Fd, Fs); break;
		case FP_MOV      : sprintf(String, "%08x d.mov    %s = %s", Offset, Fd, Fs); break;
		case FP_NEG      : sprintf(String, "%08x d.neg    %s = -%s", Offset, Fd, Fs); break;
		case FP_ROUNDL   : sprintf(String, "%08x d.roundl %s = %s", Offset, Fd, Fs); break;
		case FP_TRUNCL   : sprintf(String, "%08x d.truncl %s = %s", Offset, Fd, Fs); break;
		case FP_CEILL    : sprintf(String, "%08x d.ceill  %s = %s", Offset, Fd, Fs); break;
		case FP_FLOORL   : sprintf(String, "%08x d.floorl %s = %s", Offset, Fd, Fs); break;
		case FP_ROUNDW   : sprintf(String, "%08x d.roundw %s = %s", Offset, Fd, Fs); break;
		case FP_TRUNCW   : sprintf(String, "%08x d.truncw %s = %s", Offset, Fd, Fs); break;
		case FP_CEILW    : sprintf(String, "%08x d.ceilw  %s = %s", Offset, Fd, Fs); break;
		case FP_FLOORW   : sprintf(String, "%08x d.floorw %s = %s", Offset, Fd, Fs); break;
		case FP_CVTS     : sprintf(String, "%08x d.cvts   %s = %s", Offset, Fd, Fs); break;
		case FP_CVTW     : sprintf(String, "%08x d.cvtw   %s = %s", Offset, Fd, Fs); break;
		case FP_CVTL     : sprintf(String, "%08x d.cvtl   %s = %s", Offset, Fd, Fs); break;
		case FP_CF       : sprintf(String, "%08x d.cf     c = %s f %s", Offset, Fs, Ft); break;
		case FP_CUN      : sprintf(String, "%08x d.cun    c = %s un %s", Offset, Fs, Ft); break;
		case FP_CEQ      : sprintf(String, "%08x d.ceq    c = %s eq %s", Offset, Fs, Ft); break;
		case FP_CUEQ     : sprintf(String, "%08x d.cueq   c = %s ueq %s", Offset, Fs, Ft); break;
		case FP_COLT     : sprintf(String, "%08x d.cotl   c = %s olt %s", Offset, Fs, Ft); break;
		case FP_CULT     : sprintf(String, "%08x d.cult   c = %s ult %s", Offset, Fs, Ft); break;
		case FP_COLE     : sprintf(String, "%08x d.cole   c = %s ole %s", Offset, Fs, Ft); break;
		case FP_CULE     : sprintf(String, "%08x d.cule   c = %s ule %s", Offset, Fs, Ft); break;
		case FP_CSF      : sprintf(String, "%08x d.csf    c = %s sf %s", Offset, Fs, Ft); break;
		case FP_CNGLE    : sprintf(String, "%08x d.cngle  c = %s ngle %s", Offset, Fs, Ft); break;
		case FP_CSEQ     : sprintf(String, "%08x d.cseq   c = %s seq %s", Offset, Fs, Ft); break;
		case FP_CNGL     : sprintf(String, "%08x d.cngl   c = %s ngl %s", Offset, Fs, Ft); break;
		case FP_CLT      : sprintf(String, "%08x d.clt    c = %s lt %s", Offset, Fs, Ft); break;
		case FP_CNGE     : sprintf(String, "%08x d.cnge   c = %s nge %s", Offset, Fs, Ft); break;
		case FP_CLE      : sprintf(String, "%08x d.cf     c = %s le %s", Offset, Fs, Ft); break;
		case FP_CNGT     : sprintf(String, "%08x d.cf     c = %s ngt %s", Offset, Fs, Ft); break;
		default          : sprintf(String, "%08x Illegal Instruction", Offset); break;
	}

	return String;
}

char *MInstr_ToString_W(MInstr Instr)
{
	static char String[50];

	switch (MInstr_Function(Instr))
	{
		case FP_CVTS     : sprintf(String, "%08x w.cvts   %s = %s", Offset, Fd, Fs); break;
		case FP_CVTD     : sprintf(String, "%08x w.cvtd   %s = %s", Offset, Fd, Fs); break;
		default          : sprintf(String, "%08x Illegal Instruction", Offset); break;
	}

	return String;
}

char *MInstr_ToString_L(MInstr Instr)
{
	static char String[50];

	switch (MInstr_Function(Instr))
	{
		case FP_CVTS     : sprintf(String, "%08x l.cvts   %s = %s", Offset, Fd, Fs); break;
		case FP_CVTD     : sprintf(String, "%08x l.cvtd   %s = %s", Offset, Fd, Fs); break;
		default          : sprintf(String, "%08x Illegal Instruction", Offset); break;
	}

	return String;
}

char *MInstr_ToString_Cop1(MInstr Instr)
{
	static char String[50];
	
	switch (MInstr_Rs(Instr))
	{
		case COP1_MFC1   : sprintf(String, "%08x mfc1     %s = %s", Offset, Rt, Fs); break;
		case COP1_DMFC1  : sprintf(String, "%08x dmfc1    %s = %s", Offset, Rt, Fs); break;
		case COP1_CFC1   : sprintf(String, "%08x cfc1     %s = %s", Offset, Rt, Fs); break;
		case COP1_MTC1   : sprintf(String, "%08x mtc1     %s = %s", Offset, Fs, Rt); break;
		case COP1_DMTC1  : sprintf(String, "%08x dmtc1    %s = %s", Offset, Fs, Rt); break;
		case COP1_CTC1   : sprintf(String, "%08x ctc1     %s = %s", Offset, Fs, Rt); break;
		case COP1_BC1    : return MInstr_ToString_Bc1(Instr);
		case COP1_S      : return MInstr_ToString_S(Instr);
		case COP1_D      : return MInstr_ToString_D(Instr);
		case COP1_W      : return MInstr_ToString_W(Instr);
		case COP1_L      : return MInstr_ToString_L(Instr);
		default          : sprintf(String, "%08x Illegal Instruction", Offset); break;
	}
}

char *MInstr_ToString(MInstr Instr)
{
	static char String[50];

	switch (MInstr_Opcode(Instr))
	{
		case OP_SPECIAL : return MInstr_ToString_Special(Instr);
		case OP_REGIMM  : return MInstr_ToString_Regimm(Instr);
		case OP_J       : sprintf(String, "%08x j        ->%08x", Offset, JumpPc); break; 
		case OP_JAL     : sprintf(String, "%08x jal      ->%08x", Offset, JumpPc); break; 
		case OP_BEQ     : sprintf(String, "%08x beq      (%s == %s) ->%08x", Offset, Rs, Rt, BranchPc); break; 
		case OP_BNE     : sprintf(String, "%08x bne      (%s != %s) ->%08x", Offset, Rs, Rt, BranchPc); break;
		case OP_BLEZ    : sprintf(String, "%08x blez     (%s <= 0) ->%08x", Offset, Rs, BranchPc); break;
		case OP_BGTZ    : sprintf(String, "%08x bgtz     (%s > 0) ->%08x", Offset, Rs, BranchPc); break;
		case OP_ADDI    : sprintf(String, "%08x addi     %s = %s + %04x", Offset, Rt, Rs, Immediate); break;
		case OP_ADDIU   : sprintf(String, "%08x addiu    %s = %s + %04x", Offset, Rt, Rs, Immediate); break;
		case OP_SLTI    : sprintf(String, "%08x slti     %s = (%s < %04x)", Offset, Rt, Rs, Immediate); break;
		case OP_SLTIU   : sprintf(String, "%08x slti     %s = (%s < %04x)", Offset, Rt, Rs, Immediate); break;
		case OP_ANDI    : sprintf(String, "%08x andi     %s = %s & %04x)", Offset, Rt, Rs, Immediate); break;
		case OP_ORI     : sprintf(String, "%08x ori      %s = %s | %04x)", Offset, Rt, Rs, Immediate); break; 
		case OP_XORI    : sprintf(String, "%08x xori     %s = %s ^ %04x)", Offset, Rt, Rs, Immediate); break; 
		case OP_LUI     : sprintf(String, "%08x lui      %s = %08x)", Offset, Rt, Immediate << 16); break;
		case OP_COP0    : return MInstr_ToString_Cop0(Instr);
		case OP_COP1    : return MInstr_ToString_Cop1(Instr);
		case OP_BEQL    : sprintf(String, "%08x beql     (%s == %s) ->%08x", Offset, Rs, Rt, BranchPc); break; 
		case OP_BNEL    : sprintf(String, "%08x bnel     (%s != %s) ->%08x", Offset, Rs, Rt, BranchPc); break;
		case OP_BLEZL   : sprintf(String, "%08x blezl    (%s <= 0) ->%08x", Offset, Rs, BranchPc); break;
		case OP_BGTZL   : sprintf(String, "%08x bgtzl    (%s > 0) ->%08x", Offset, Rs, BranchPc); break;
		case OP_DADDI   : sprintf(String, "%08x daddi    %s = %s + %04x", Offset, Rt, Rs, Immediate); break;
		case OP_DADDIU  : sprintf(String, "%08x daddiu   %s = %s + %04x", Offset, Rt, Rs, Immediate); break;
		case OP_LDL     : sprintf(String, "%08x ldl      %s = [%s + %04x]", Offset, Rt, Rs, Immediate); break;
		case OP_LDR     : sprintf(String, "%08x ldr      %s = [%s + %04x]", Offset, Rt, Rs, Immediate); break;
		case OP_LB      : sprintf(String, "%08x lb       %s = [%s + %04x]", Offset, Rt, Rs, Immediate); break;
		case OP_LH      : sprintf(String, "%08x lh       %s = [%s + %04x]", Offset, Rt, Rs, Immediate); break;
		case OP_LWL     : sprintf(String, "%08x lwl      %s = [%s + %04x]", Offset, Rt, Rs, Immediate); break;
		case OP_LW      : sprintf(String, "%08x lw       %s = [%s + %04x]", Offset, Rt, Rs, Immediate); break;
		case OP_LBU     : sprintf(String, "%08x lbu      %s = [%s + %04x]", Offset, Rt, Rs, Immediate); break;
		case OP_LHU     : sprintf(String, "%08x lhu      %s = [%s + %04x]", Offset, Rt, Rs, Immediate); break;
		case OP_LWR     : sprintf(String, "%08x lwr      %s = [%s + %04x]", Offset, Rt, Rs, Immediate); break;
		case OP_LWU     : sprintf(String, "%08x lwu      %s = [%s + %04x]", Offset, Rt, Rs, Immediate); break;
		case OP_SB      : sprintf(String, "%08x sb       [%s + %04x] = %s", Offset, Rs, Immediate, Rt); break;
		case OP_SH      : sprintf(String, "%08x sh       [%s + %04x] = %s", Offset, Rs, Immediate, Rt); break; 
		case OP_SWL     : sprintf(String, "%08x swl      [%s + %04x] = %s", Offset, Rs, Immediate, Rt); break; 
		case OP_SW      : sprintf(String, "%08x sw       [%s + %04x] = %s", Offset, Rs, Immediate, Rt); break; 
		case OP_SDL     : sprintf(String, "%08x sdl      [%s + %04x] = %s", Offset, Rs, Immediate, Rt); break; 
		case OP_SDR     : sprintf(String, "%08x sdr      [%s + %04x] = %s", Offset, Rs, Immediate, Rt); break; 
		case OP_SWR     : sprintf(String, "%08x swr      [%s + %04x] = %s", Offset, Rs, Immediate, Rt); break;
		case OP_CACHE   : sprintf(String, "%08x cache", Offset); break;
		case OP_LL      : sprintf(String, "%08x ll       %s = [%s + %04x]", Offset, Rt, Rs, Immediate); break;
		case OP_LWC1    : sprintf(String, "%08x lwc1     %s = [%s + %04x]", Offset, Ft, Rs, Immediate); break;
		case OP_LLD     : sprintf(String, "%08x lld      %s = [%s + %04x]", Offset, Rt, Rs, Immediate); break;
		case OP_LDC1    : sprintf(String, "%08x ldc1     %s = [%s + %04x]", Offset, Ft, Rs, Immediate); break;  
		case OP_LDC2    : sprintf(String, "%08x ldc2", Offset); break;
		case OP_LD      : sprintf(String, "%08x ld       %s = [%s + %04x]", Offset, Rt, Rs, Immediate); break;
		case OP_SC      : sprintf(String, "%08x sc       [%s + %04x] = %s", Offset, Rs, Immediate, Rt); break;
		case OP_SWC1    : sprintf(String, "%08x swc1     [%s + %04x] = %s", Offset, Rs, Immediate, Ft); break;
		case OP_SCD     : sprintf(String, "%08x scd      [%s + %04x] = %s", Offset, Rs, Immediate, Rt); break; 
		case OP_SDC1    : sprintf(String, "%08x sdc1     [%s + %04x] = %s", Offset, Rs, Immediate, Ft); break;
		case OP_SDC2    : sprintf(String, "%08x sdc2", Offset); break;
		case OP_SD      : sprintf(String, "%08x sd       [%s + %04x] = %s", Offset, Rs, Immediate, Rt); break;
		default         : sprintf(String, "%08x Illegal Instruction", Offset); break;
	}

	return String;
}
